@@ -4,6 +4,8 @@
 import argparse
 import copy
 import logging
+import os
+import os.path
 import stat
 import sys
 from collections.abc import MutableSequence, Sequence
@@ -62,6 +64,8 @@ def main(args: Optional[List[str]] = None) -> int:
 def run(args: argparse.Namespace) -> int:
     """Main function."""
     imports: Set[str] = set()
+    if args.dir and not os.path.exists(args.dir):
+        os.makedirs(args.dir)
     for path in args.inputs:
         _logger.info("Processing %s", path)
         document = load_cwl_document(path)
@@ -76,45 +80,91 @@ def run(args: argparse.Namespace) -> int:
                 if args.v1_1_only:
                     _logger.info("Skipping v1.1 document as requested: %s.", path)
                     continue
+
+            if args.v1_only:
+                target_version = "v1.0"
+            elif args.v1_1_only:
+                target_version = "v1.1"
+            else:
+                target_version = "latest"
             upgraded_document = upgrade_document(
-                document, args.v1_only, args.v1_1_only, args.dir, imports
+                document,
+                args.dir,
+                target_version=target_version,
+                imports=imports,
             )
-            write_cwl_document(upgraded_document, Path(path).name, args.dir)
+            if upgraded_document is not None:
+                write_cwl_document(upgraded_document, Path(path).name, args.dir)
     return 0
 
 
 def upgrade_document(
-    document: Any, v1_only: bool, v1_1_only: bool, output_dir: str, imports: Set[str]
+    document: Any,
+    output_dir: str,
+    target_version: Optional[str] = "latest",
+    imports: Optional[Set[str]] = None,
 ) -> Any:
+    if imports is None:
+        imports = set()
+    supported_versions = ["v1.0", "v1.1", "v1.2", "latest"]
+    if target_version not in supported_versions:
+        _logger.error(f"Unsupported target cwlVersion: {target_version}")
+        return
+
     version = document["cwlVersion"]
+    main_updater = None
+    inner_updater = None
+
     if version == "cwl:draft-3" or version == "draft-3":
-        if v1_only:
+        if target_version == "v1.0":
             main_updater = draft3_to_v1_0
             inner_updater = _draft3_to_v1_0
-        elif v1_1_only:
+        elif target_version == "v1.1":
             main_updater = draft3_to_v1_1
             inner_updater = _draft3_to_v1_1
-        else:
+        elif target_version == "v1.2":
+            main_updater = draft3_to_v1_2
+            inner_updater = _draft3_to_v1_2
+        elif target_version == "latest":
             main_updater = draft3_to_v1_2
             inner_updater = _draft3_to_v1_2
     elif version == "v1.0":
-        if v1_only:
-            _logger.info("Skipping v1.0 document as requested.")
+        if target_version == "v1.0":
+            _logger.info("Not upgrading v1.0 document as requested.")
             return
-        elif v1_1_only:
+        elif target_version == "v1.1":
             main_updater = v1_0_to_v1_1
             inner_updater = _v1_0_to_v1_1
-        else:
+        elif target_version == "v1.2":
+            main_updater = v1_0_to_v1_2
+            inner_updater = _v1_0_to_v1_2
+        elif target_version == "latest":
             main_updater = v1_0_to_v1_2
             inner_updater = _v1_0_to_v1_2
     elif version == "v1.1":
-        if v1_1_only:
-            _logger.info("Skipping v1.1 document as requested.")
+        if target_version == "v1.1":
+            _logger.info("Not upgrading v1.1 document as requested.")
+            return
+        elif target_version == "v1.2":
+            main_updater = v1_1_to_v1_2
+            inner_updater = _v1_1_to_v1_2
+        elif target_version == "latest":
+            main_updater = v1_1_to_v1_2
+            inner_updater = _v1_1_to_v1_2
+    elif version == "v1.2":
+        if target_version == "v1.2":
+            _logger.info("Not upgrading v1.2 document as requested.")
+            return
+        elif target_version == "latest":
             return
-        main_updater = v1_1_to_v1_2
-        inner_updater = _v1_1_to_v1_2
     else:
-        _logger.error(f"Unsupported cwlVersion: {version}")
+        _logger.error(f"Unknown cwlVersion in source document: {version}")
+        return
+
+    if main_updater is None or inner_updater is None:
+        _logger.error(f"Cannot downgrade from cwlVersion {version} to {target_version}")
+        return
+
     process_imports(document, imports, inner_updater, output_dir)
     return main_updater(document, output_dir)
 
@@ -178,15 +228,14 @@ def process_imports(
 def v1_0_to_v1_1(document: CommentedMap, outdir: str) -> CommentedMap:
     """CWL v1.0.x to v1.1 transformation loop."""
     _v1_0_to_v1_1(document, outdir)
-    if isinstance(document, MutableMapping):
-        for key, value in document.items():
-            with SourceLine(document, key, Exception):
-                if isinstance(value, CommentedMap):
-                    document[key] = _v1_0_to_v1_1(value, outdir)
-                elif isinstance(value, list):
-                    for index, entry in enumerate(value):
-                        if isinstance(entry, CommentedMap):
-                            value[index] = _v1_0_to_v1_1(entry, outdir)
+    for key, value in document.items():
+        with SourceLine(document, key, Exception):
+            if isinstance(value, CommentedMap):
+                document[key] = _v1_0_to_v1_1(value, outdir)
+            elif isinstance(value, list):
+                for index, entry in enumerate(value):
+                    if isinstance(entry, CommentedMap):
+                        value[index] = _v1_0_to_v1_1(entry, outdir)
     document["cwlVersion"] = "v1.1"
     return sort_v1_0(document)
 
@@ -306,12 +355,17 @@ def _v1_0_to_v1_1(document: CommentedMap, outdir: str) -> CommentedMap:
                                 _v1_0_to_v1_1(process, outdir)
                                 if "cwlVersion" in process:
                                     del process["cwlVersion"]
-                            elif isinstance(entry["run"], str):
+                            elif (
+                                isinstance(entry["run"], str)
+                                and "#" not in entry["run"]
+                            ):
                                 path = Path(document.lc.filename).parent / entry["run"]
                                 process = v1_0_to_v1_1(
                                     load_cwl_document(str(path)), outdir
                                 )
                                 write_cwl_document(process, path.name, outdir)
+                            elif isinstance(entry["run"], str) and "#" in entry["run"]:
+                                pass  # reference to $graph entry
                             else:
                                 raise Exception(
                                     "'run' entry was neither a CWL Process nor "
@@ -323,13 +377,13 @@ def _v1_0_to_v1_1(document: CommentedMap, outdir: str) -> CommentedMap:
             move_up_loadcontents(document)
             network_access = has_hint_or_req(document, "NetworkAccess")
             listing = has_hint_or_req(document, "LoadListingRequirement")
-            hints = document.get("hints", {})
+            reqs = document.get("requirements", {})
             # TODO: add comments to explain the extra hints
-            if isinstance(hints, MutableSequence):
+            if isinstance(reqs, MutableSequence):
                 if not network_access:
-                    hints.append({"class": "NetworkAccess", "networkAccess": True})
+                    reqs.append({"class": "NetworkAccess", "networkAccess": True})
                 if not listing:
-                    hints.append(
+                    reqs.append(
                         cmap(
                             {
                                 "class": "LoadListingRequirement",
@@ -337,15 +391,15 @@ def _v1_0_to_v1_1(document: CommentedMap, outdir: str) -> CommentedMap:
                             }
                         )
                     )
-            elif isinstance(hints, MutableMapping):
+            elif isinstance(reqs, MutableMapping):
                 if not network_access:
-                    hints["NetworkAccess"] = {"networkAccess": True}
+                    reqs["NetworkAccess"] = {"networkAccess": True}
                 if not listing:
-                    hints["LoadListingRequirement"] = cmap(
+                    reqs["LoadListingRequirement"] = cmap(
                         {"loadListing": "deep_listing"}
                     )
-            if "hints" not in document:
-                document["hints"] = hints
+            if "requirements" not in document:
+                document["requirements"] = reqs
         elif document["class"] == "ExpressionTool":
             move_up_loadcontents(document)
             cleanup_v1_0_input_bindings(document)
@@ -543,18 +597,20 @@ def input_output_clean(document: Dict[str, Any]) -> None:
                 if "$import" in param:
                     meta = True
         if not meta:
-            for index2, param in enumerate(document[param_type]):
+            for index2, param2 in enumerate(document[param_type]):
                 with SourceLine(document[param_type], index2, Exception):
-                    param_id = param.pop("id").lstrip("#")
-                    if "type" in param:
-                        param["type"] = shorten_type(param["type"])
-                        array_type_raise_sf(param)
-                    if "description" in param:
-                        param["doc"] = param.pop("description")
-                    if len(param) > 1:
-                        new_section[param_id] = sort_input_or_output(param)
+                    param_id = param2.pop("id").lstrip("#")
+                    if "type" in param2:
+                        param2["type"] = shorten_type(param2["type"])
+                        array_type_raise_sf(param2)
+                    if "description" in param2:
+                        param2["doc"] = param2.pop("description")
+                    if len(param2) > 1:
+                        new_section[param_id] = sort_input_or_output(param2)
+                    elif "type" in param2 and isinstance(param2["type"], str):
+                        new_section[param_id] = param2.popitem()[1]
                     else:
-                        new_section[param_id] = param.popitem()[1]
+                        new_section[param_id] = param2
             document[param_type] = new_section
 
 
